/*
NSIM is a simulator for the processor used in the HP-41 (Nut) and in the HP
Series 10 (Voyager) calculators.

Copyright 1995 Eric L. Smith

NSIM is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License version 2 as published by the Free
Software Foundation.  Note that I am not granting permission to redistribute
or modify NSIM under the terms of any later version of the General Public
License.

This program is distributed in the hope that it will be useful (or at least
amusing), but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along with
this program (in the file "COPYING"); if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

$Header: /home/yl2/eric/hpcalc/nasm/RCS/xio.c,v 1.9 1995/08/01 15:08:00 eric Exp eric $
*/

#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "xio.h"


Display *mydisplay;
int myscreen;
Window mywindow;
GC mygc;
XFontStruct *myfont;

/* pixel values */
unsigned long black, white;

int window_width, window_height;

char disp_buf1 [40];
char disp_buf2 [40];

int pressed_key = -1;


#define MAX_KEY 256
#define ESCAPE_CODE -2
static int keymap [MAX_KEY];

static void init_keymap (void)
{
  int i;

  for (i = 0; i < MAX_KEY; i++)
    keymap [i] = -1;

  keymap [XKeysymToKeycode (mydisplay, XK_Escape)] = ESCAPE_CODE;

  keymap [XKeysymToKeycode (mydisplay, XK_a)] = 11;
  keymap [XKeysymToKeycode (mydisplay, XK_b)] = 12;
  keymap [XKeysymToKeycode (mydisplay, XK_c)] = 13;
  keymap [XKeysymToKeycode (mydisplay, XK_d)] = 14;
  keymap [XKeysymToKeycode (mydisplay, XK_e)] = 15;

  keymap [XKeysymToKeycode (mydisplay, XK_f)] = 21;
  keymap [XKeysymToKeycode (mydisplay, XK_g)] = 22;
  keymap [XKeysymToKeycode (mydisplay, XK_h)] = 23;
  keymap [XKeysymToKeycode (mydisplay, XK_i)] = 24;
  keymap [XKeysymToKeycode (mydisplay, XK_j)] = 25;

  keymap [XKeysymToKeycode (mydisplay, XK_k)] = 32;
  keymap [XKeysymToKeycode (mydisplay, XK_l)] = 33;
  keymap [XKeysymToKeycode (mydisplay, XK_m)] = 34;

  keymap [XKeysymToKeycode (mydisplay, XK_n)] = 41;
  keymap [XKeysymToKeycode (mydisplay, XK_o)] = 43;
  keymap [XKeysymToKeycode (mydisplay, XK_p)] = 44;
  keymap [XKeysymToKeycode (mydisplay, XK_Return)] = 41;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Enter)] = 41;
  keymap [XKeysymToKeycode (mydisplay, XK_BackSpace)] = 45;
  keymap [XKeysymToKeycode (mydisplay, XK_Delete)] = 45;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Delete)] = 45;

  keymap [XKeysymToKeycode (mydisplay, XK_q)] = 51;
  keymap [XKeysymToKeycode (mydisplay, XK_r)] = 52;
  keymap [XKeysymToKeycode (mydisplay, XK_s)] = 53;
  keymap [XKeysymToKeycode (mydisplay, XK_t)] = 54;
  keymap [XKeysymToKeycode (mydisplay, XK_minus)] = 51;
  keymap [XKeysymToKeycode (mydisplay, XK_7)] = 52;
  keymap [XKeysymToKeycode (mydisplay, XK_8)] = 53;
  keymap [XKeysymToKeycode (mydisplay, XK_9)] = 54;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Subtract)] = 51;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_7)] = 52;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_8)] = 53;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_9)] = 54;

  keymap [XKeysymToKeycode (mydisplay, XK_u)] = 61;
  keymap [XKeysymToKeycode (mydisplay, XK_v)] = 62;
  keymap [XKeysymToKeycode (mydisplay, XK_w)] = 63;
  keymap [XKeysymToKeycode (mydisplay, XK_x)] = 64;
  keymap [XKeysymToKeycode (mydisplay, XK_plus)] = 61;
  keymap [XKeysymToKeycode (mydisplay, XK_4)] = 62;
  keymap [XKeysymToKeycode (mydisplay, XK_5)] = 63;
  keymap [XKeysymToKeycode (mydisplay, XK_6)] = 64;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Add)] = 61;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_4)] = 62;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_5)] = 63;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_6)] = 64;

  keymap [XKeysymToKeycode (mydisplay, XK_y)] = 71;
  keymap [XKeysymToKeycode (mydisplay, XK_z)] = 72;
  keymap [XKeysymToKeycode (mydisplay, XK_equal)] = 73;
  keymap [XKeysymToKeycode (mydisplay, XK_question)] = 74;
  keymap [XKeysymToKeycode (mydisplay, XK_asterisk)] = 71;
  keymap [XKeysymToKeycode (mydisplay, XK_1)] = 72;
  keymap [XKeysymToKeycode (mydisplay, XK_2)] = 73;
  keymap [XKeysymToKeycode (mydisplay, XK_3)] = 74;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Multiply)] = 71;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_1)] = 72;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_2)] = 73;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_3)] = 74;

  keymap [XKeysymToKeycode (mydisplay, XK_colon)] = 81;
  keymap [XKeysymToKeycode (mydisplay, XK_space)] = 82;
  keymap [XKeysymToKeycode (mydisplay, XK_comma)] = 83;
  keymap [XKeysymToKeycode (mydisplay, XK_slash)] = 81;
  keymap [XKeysymToKeycode (mydisplay, XK_0)] = 82;
  keymap [XKeysymToKeycode (mydisplay, XK_period)] = 83;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Divide)] = 81;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_0)] = 82;
  keymap [XKeysymToKeycode (mydisplay, XK_KP_Decimal)] = 83;
}

#ifdef SMALL
#define DISPLAY_HEIGHT 48
#define VERT_SPACING 40
#else
#define DISPLAY_HEIGHT 72
#define VERT_SPACING 48
#endif

#define UC1 24
#define UC2 72
#define UC3 120
#define UC4 168
#define UC5 216

#define LC1 24
#define LC2 73
#define LC3 141
#define LC4 209

#define WINDOW_WIDTH  270

#define R0 (DISPLAY_HEIGHT + 8)
#define R1 (DISPLAY_HEIGHT + 1 * VERT_SPACING)
#define R2 (DISPLAY_HEIGHT + 2 * VERT_SPACING)
#define R3 (DISPLAY_HEIGHT + 3 * VERT_SPACING)
#define R4 (DISPLAY_HEIGHT + 4 * VERT_SPACING)
#define R5 (DISPLAY_HEIGHT + 5 * VERT_SPACING)
#define R6 (DISPLAY_HEIGHT + 6 * VERT_SPACING)
#define R7 (DISPLAY_HEIGHT + 7 * VERT_SPACING)
#define R8 (DISPLAY_HEIGHT + 8 * VERT_SPACING)

#define WINDOW_HEIGHT (DISPLAY_HEIGHT + 9 * VERT_SPACING)

XRectangle display_rect1 = { 0, 0, WINDOW_WIDTH, DISPLAY_HEIGHT / 2 };
XRectangle display_rect2 = { 0, DISPLAY_HEIGHT / 2, WINDOW_WIDTH, DISPLAY_HEIGHT / 2 };


#define BLACK    0
#define WHITE    1
#define DK_GREY  2
#define GOLD     3
#define BLUE     4
#define DISPLAY  5

#define MAX_COL  6


unsigned long pixval [MAX_COL];

typedef struct
{
  unsigned short red;
  unsigned short green;
  unsigned short blue;
} rgb;


rgb color [MAX_COL] =
{
  { 0x0000, 0x0000, 0x0000 },
  { 0xffff, 0xffff, 0xffff },
  { 0x3800, 0x3800, 0x3800 },
  { 0xf7f7, 0xb2b2, 0x3434 },
  { 0xa000, 0xa000, 0xffff },
  { 0x8e8e, 0x7d7d, 0x6d6d },
};

static void init_colors (void)
{
  int i;
  XColor col;
  Colormap cmap;

  cmap = DefaultColormap (mydisplay, myscreen);

  for (i = 0; i < MAX_COL; i++)
    {
      col.red   = color [i].red;
      col.green = color [i].green;
      col.blue  = color [i].blue;
      if (XAllocColor (mydisplay, cmap, & col) == 0)
	{
	  fprintf (stderr, "can't get color\n");
	  exit (2);
	}
      pixval [i] = col.pixel;
    }
}


typedef struct
{
  XRectangle rect;
  char *label;
  char *flabel;
  char *alabel;
  int keycode;
  int fg, bg;
} keyinfo;

#define S_SPCL 30,16 /* special keys (ON, USER, PRGM, ALPHA) */
#define S_NORM 30,24 /* normal keys */
#define S_DBL  78,24 /* ENTER key */
#define S_ARTH 24,24 /* arithmetic keys */
#define S_NUM  37,24 /* numeric keys */

keyinfo keys [] =
{
  { { UC1, R0, S_SPCL }, "ON",    "",        "",  01, WHITE, BLACK },
  { { UC2, R0, S_SPCL }, "USER",  "",        "",  02, WHITE, BLACK },
  { { UC4, R0, S_SPCL }, "PRGM",  "",        "",  03, WHITE, BLACK },
  { { UC5, R0, S_SPCL }, "ALPHA", "",        "",  04, WHITE, BLACK },

  { { UC1, R1, S_NORM }, "SIG+",  "SIG-",    "A", 11, WHITE, BLACK },
  { { UC2, R1, S_NORM }, "1/x",   "y^x",     "B", 12, WHITE, BLACK },
  { { UC3, R1, S_NORM }, "sqr(x)","x^2",     "C", 13, WHITE, BLACK },
  { { UC4, R1, S_NORM }, "LOG",   "10^x",    "D", 14, WHITE, BLACK },
  { { UC5, R1, S_NORM }, "LN",    "e^x",     "E", 15, WHITE, BLACK },

  { { UC1, R2, S_NORM }, "x<>y",  "CLSIG",   "F", 21, WHITE, BLACK },
  { { UC2, R2, S_NORM }, "RDN",   "%",       "G", 22, WHITE, BLACK },
  { { UC3, R2, S_NORM }, "SIN",   "SIN^-1",  "H", 23, WHITE, BLACK },
  { { UC4, R2, S_NORM }, "COS",   "COS^-1",  "I", 24, WHITE, BLACK },
  { { UC5, R2, S_NORM }, "TAN",   "TAN^-1",  "J", 25, WHITE, BLACK },

  { { UC1, R3, S_NORM }, "",      "",        "",  31, GOLD,  GOLD },
  { { UC2, R3, S_NORM }, "XEQ",   "ASN",     "K", 32, WHITE, BLACK },
  { { UC3, R3, S_NORM }, "STO",   "LBL"   ,  "L", 33, WHITE, BLACK },
  { { UC4, R3, S_NORM }, "RCL",   "GTO",     "M", 34, WHITE, BLACK },
  { { UC5, R3, S_NORM }, "SST",   "BST",     "",  35, WHITE, BLACK },

  { { UC1, R4, S_DBL  }, "ENTER^","CATALOG", "N", 41, WHITE, BLACK },
  { { UC3, R4, S_NORM }, "CHS",   "ISG",     "O", 43, WHITE, BLACK },
  { { UC4, R4, S_NORM }, "EEX",   "RTN",     "P", 44, WHITE, BLACK },
  { { UC5, R4, S_NORM }, "<--",   "CLX/A",   "",  45, WHITE, BLACK },

  { { LC1, R5, S_ARTH }, "-",     "x=y?",    "Q", 51, WHITE, BLACK },
  { { LC2, R5, S_NUM  }, "7",     "SF",      "R", 52, WHITE, BLACK },
  { { LC3, R5, S_NUM  }, "8",     "CF",      "S", 53, WHITE, BLACK },
  { { LC4, R5, S_NUM  }, "9",     "FS?",     "T", 54, WHITE, BLACK },

  { { LC1, R6, S_ARTH }, "+",     "x<=y?",   "U", 61, WHITE, BLACK },
  { { LC2, R6, S_NUM  }, "4",     "BEEP",    "V", 62, WHITE, BLACK },
  { { LC3, R6, S_NUM  }, "5",     "P->R",    "W", 63, WHITE, BLACK },
  { { LC4, R6, S_NUM  }, "6",     "R->P",    "X", 64, WHITE, BLACK },

  { { LC1, R7, S_ARTH }, "x",     "x>y?",    "Y", 71, WHITE, BLACK },
  { { LC2, R7, S_NUM  }, "1",     "FIX",     "Z", 72, WHITE, BLACK },
  { { LC3, R7, S_NUM  }, "2",     "SCI",     "=", 73, WHITE, BLACK },
  { { LC4, R7, S_NUM  }, "3",     "ENG",     "?", 74, WHITE, BLACK },

  { { LC1, R8, S_ARTH }, "/",     "x=0?",    ":", 81, WHITE, BLACK },
  { { LC2, R8, S_NUM  }, "0",     "Pi",      "SPACE", 82, WHITE, BLACK },
  { { LC3, R8, S_NUM  }, ".",     "LASTx",   ",", 83, WHITE, BLACK },
  { { LC4, R8, S_NUM  }, "R/S",   "VIEW",    "",  84, WHITE, BLACK },
};


static void draw_string (char *s, XRectangle *rect, int fg, int bg)
{
  XCharStruct cs;
  int dir_hint, ascent, descent;

  XSetForeground (mydisplay, mygc, pixval [bg]);

  XFillRectangles (mydisplay, mywindow, mygc, rect, 1);

  if (s)
    {
      XTextExtents (myfont, s, strlen (s),
		    & dir_hint, & ascent, & descent, & cs);

      XSetForeground (mydisplay, mygc, pixval [fg]);
      XSetBackground (mydisplay, mygc, pixval [bg]);

      XDrawImageString (mydisplay, mywindow, mygc,
		   rect->x + (rect->width - (cs.lbearing + cs.rbearing)) / 2,
		   rect->y + rect->height - (rect->height - (cs.ascent + cs.descent)) / 2,
		   s, strlen (s));
    }
}


#define DIGIT_WIDTH 14
#define SEGLEN 8

XRectangle segrect [8] =
{
  {  0,  0,  8,  0 }, /* a */
  {  8,  0,  0,  8 }, /* b */
  {  8,  8,  0,  8 }, /* c */
  {  0, 16,  8,  0 }, /* d */
  {  0,  8,  0,  8 }, /* e */
  {  0,  0,  0,  8 }, /* f */
  {  0,  8,  8,  0 }, /* g */
  {  3, 11,  2,  2 }  /* . */
};

char segments [13] =
{ 
  0x3f, /* 0 */
  0x06, /* 1 */
  0x5b, /* 2 */
  0x4f, /* 3 */
  0x66, /* 4 */
  0x6d, /* 5 */
  0x7d, /* 6 */
  0x07, /* 7 */
  0x7f, /* 8 */
  0x6f, /* 9 */
  0x40, /* - */
  0x80, /* . */
  0x00  /*   */
};


static void draw_display (char *str1, char *str2)
{
  draw_string (str1, & display_rect1, BLACK, DISPLAY);
  draw_string (str2, & display_rect2, BLACK, DISPLAY);
  XSync (mydisplay, 0);
}


static void draw_key (int i)
{
  XRectangle r;

  r = keys [i].rect;
  if (i >= 4)
    r.height >>= 1;
  draw_string (keys [i].label, & r, keys [i].fg, keys [i].bg);
  if (i >= 4)
    {
      r.y += r.height;
      draw_string (keys [i].alabel, & r, BLUE, keys [i].bg);
    }
  r = keys [i].rect;
#ifdef SMALL
  r.y -= r.height / 2;
  r.height /= 2;
#else
  r.y -= r.height;
#endif
  draw_string (keys [i].flabel, & r, GOLD, DK_GREY);
}

static void draw_calc (void)
{
  int i;

  for (i = 0; i < (sizeof (keys) / sizeof (keyinfo)); i++)
    {
      draw_key (i);
    }
  XSync (mydisplay, 0);
}


static int pt_in_rect (int x, int y, XRectangle *rect)
{
  return ((x >= rect->x) && (x < (rect->x + rect->width)) &&
	  (y >= rect->y) && (y < (rect->y + rect->height)));
}


static int find_key (int x, int y)
{
  int i;

  for (i = 0; i < (sizeof (keys) / sizeof (keyinfo)); i++)
    {
      if (pt_in_rect (x, y, & keys [i].rect))
	return (i);
    }
  return (-1);
}


#define LINE_WIDTH 1

static void init_graphics (int argc, char *argv[], char *window_name)
{
  XSizeHints myhint;

  mydisplay = XOpenDisplay ("");
  if (!mydisplay)
    {
      fprintf (stderr, "Can't init X\n");
      exit (1);
    }

  myscreen = DefaultScreen (mydisplay);
  white = WhitePixel (mydisplay, myscreen);
  black = BlackPixel (mydisplay, myscreen);

  init_colors ();

  window_width = WINDOW_WIDTH;
  window_height = WINDOW_HEIGHT;

  myhint.x = 50; myhint.y = 50;
  myhint.width = window_width;
  myhint.height = window_height;
  myhint.flags = PPosition | PSize;

  mywindow = XCreateSimpleWindow
    (mydisplay,
     DefaultRootWindow (mydisplay),
     myhint.x, myhint.y, myhint.width, myhint.height,
     5, 
     black, pixval [DK_GREY]);

  XSetStandardProperties (mydisplay, mywindow, window_name, window_name,
			  None, argv, argc, &myhint);

  mygc = XCreateGC (mydisplay, mywindow, 0, 0);

  myfont = XLoadQueryFont (mydisplay, "6x13");
  XSetFont (mydisplay, mygc, myfont->fid);

  XSetLineAttributes (mydisplay, mygc, LINE_WIDTH, LineSolid, CapButt, JoinMiter);

  XSelectInput (mydisplay, mywindow,
		KeyPressMask | KeyReleaseMask | 
		ButtonPressMask | ButtonReleaseMask |
		ExposureMask );

  XMapRaised (mydisplay, mywindow);
}


static int handle_events (void)
{
  XEvent event;
  int log_keycode;
  int key_rect_index;

  while (XEventsQueued (mydisplay, QueuedAfterReading) != 0)
    {
      XNextEvent (mydisplay, & event);
      switch (event.type)
	{
	case KeyPress:
	  log_keycode = keymap [event.xkey.keycode];
	  if (log_keycode == ESCAPE_CODE)
	    return (1);
	  if (log_keycode >= 0)
	    pressed_key = log_keycode;
	  break;
	case KeyRelease:
	  log_keycode = keymap [event.xkey.keycode];
	  if (log_keycode == pressed_key)
	    pressed_key = -1;
	  break;
	case ButtonPress:
	  key_rect_index = find_key (event.xbutton.x, event.xbutton.y);
	  if (key_rect_index >= 0)
	    pressed_key = keys [key_rect_index].keycode;
	  break;
	case ButtonRelease:
	  key_rect_index = find_key (event.xbutton.x, event.xbutton.y);
	  if (key_rect_index >= 0)
	    {
	      log_keycode = keys [key_rect_index].keycode;
	      if (log_keycode == pressed_key)
		pressed_key = -1;
	    }
	  break;
	case Expose:
	  if (event.xexpose.count == 0)
	    {
	      draw_calc ();
	      draw_display (disp_buf1, disp_buf2);
	    }
	  break;
	case NoExpose:
	  break;
	default:
	  printf ("unknown event %d\n", event.type);
	  break;
	}
    }
  return (0);
}


void init_user_io (int argc, char *argv[])
{
  disp_buf1 [0] ='\0';
  disp_buf2 [0] ='\0';
  init_graphics (argc, argv, "nsim");
  init_keymap ();
  draw_calc ();
}


void update_display (char *str1, char *str2)
{
  if ((strcmp (disp_buf1, str1) != 0) ||
      (strcmp (disp_buf2, str2) != 0))
    {
      strcpy (disp_buf1, str1);
      strcpy (disp_buf2, str2);
      draw_display (str1, str2);
    }
}


/* returns KC_NONE if no key pressed */
/* returns KC_EXIT for exit */
int check_keyboard (void)
{
  if (handle_events ())
    return (KC_EXIT);
  if (pressed_key >= 0)
    return (pressed_key);
  else
    return (KC_NONE);
}


